/*
 * KINE.H
 * 
 * Manipulation of moving points and kinetic data structures
 *
 */

#ifndef _KINE_H
#define _KINE_H

#include "btree.h"
#include <GL/gl.h>
#include <GL/glu.h>

/* inner nodes of tournament tree */

struct innerNode {

  REAL T;
  int ind;
  short un;
};


/* for debugging - just ignore it */

struct savedObj {
  P2 currP;
  int index, inCone[3],hitUp[3],hitDown[3];
};

struct savedObjs {

  int nofObjs;
  savedObj **mos;
  int cind1, cind2;
  short et, ax;
};

class mObj2D;

class eventQueue1;

class eventQueue2;

class eventQueue3;

/* moving point in 1 dimension */

class mObj1D {

public:

  // start coordinate of the point
  REAL startpoint;

  // speed of the point
  REAL speed;

  // start time
  REAL starttime;

  // the 2D point it is related - 1D moving point is obtained by 
  // projecting 2D point along some specific axis.
  mObj2D *org;

  mObj1D(void) { starttime = 0; }
  mObj1D(REAL,REAL);
  mObj1D(REAL,REAL,REAL);
  mObj1D(mObj2D*);

  void set(REAL,REAL,REAL);
  void init(mObj2D*);

  update(REAL,REAL,REAL);
  update(mObj1D);

  REAL meet(mObj1D,REAL);

  REAL pos(REAL t) { return (t<starttime)?(-MAX):
    (startpoint+(t-starttime)*speed);}

  REAL distance(mObj1D,REAL);

  short order(mObj1D*,REAL);

  void draw(REAL);
};

/* flightplan of 2D point */

class flightPlan2 {
  
public:
  
  P2 startpoint;
  P2 speed;
  REAL starttime;

  flightPlan2(void) { starttime = 0;}
  flightPlan2(P2,P2);  
  flightPlan2(P2,P2,REAL);

  update(P2,P2,REAL);
  update(flightPlan2);

  void makeorder(mObj1D*,short);
  short pos(P2, REAL);
  bounced(REAL,REAL,short);
  REAL hitBoundary(Boundary *, REAL, short *);
};

/* 2D points in motion */

class mObj2D {

public:

  // radius of the ball
  REAL radius;
  
  P2 normvec;
  REAL thetastep;

  // index of the point - each object has an unique index
  int index;

  // flight plan of the point
  flightPlan2 fp;

  // projection of the motion on 3 directions
  mObj1D ordaxis[3];

  // candidate trees of 3 directions
  bbTree cand[3];

  // upper parent trees of 3 directions
  bbTree parentU[3];

  // lower parent trees of 3 directions
  bbTree parentL[3];

  /* the related node of the point
     relNode: nodes in candidate trees
     cuNode:  nodes in parentU trees
     clNode:  nodes in parentL trees
     */
  bbTNode relNode[3];
  bbTNode cuNode[3];
  bbTNode clNode[3];

  // order of the point along 3 directions
  int ind_q1[3];

  // leftmost candidate point in the cone
  mObj2D *lcand[3];

  mObj2D(void);
  mObj2D(flightPlan2, REAL);
  mObj2D(flightPlan2);
  mObj2D(P2, P2);
  mObj2D(P2, P2, REAL);

  void set(P2,P2,REAL,REAL);

  void resetNodes(void);
  void resetNodes(REAL);

  update(flightPlan2);
  update(P2,P2,REAL);

  inline short pos(P2 p, REAL t) { return fp.pos(p,t); }

  void draw(REAL);

  REAL distance2(mObj2D);
  REAL distance2(mObj2D, REAL);
  REAL closestcand(mObj2D **, REAL);

  inline REAL ord(short axis) { 
    short ax,si;
    ax = axis%3;
    si = axis/3;
    if(ax<0) ax+=3;
    return (si%2)?(ind_q1[ax]):(-ind_q1[ax]);
  }
    
  REAL sord(short);
  inline void setindex(int i) {index = i;}

  // if this point is in the same cone of direction axis with p, return 
  // the point related to the cone, return NULL otherwise
  mObj2D* inSameCone(mObj2D *p, short axis);

  // return the point of which this point lies in the cone. 
  mObj2D* inTree1(short axis);

  // return the point of which this point hit the upper or lower edges
  mObj2D* inTree2(short,short);

  // recompute the trajectory of the point after bounced with wall
  REAL hitBoundary(Boundary *, REAL, short *);
  bounced(Boundary *, REAL, short);

  // recommpute the trajectories after two points collide
  bounced(mObj2D *, REAL);
  
  // determine if this point is some other's leftmost candidate
  mObj2D* islcand(short);

  ~mObj2D(void);
  savedObj* save(REAL);

};


// compute when two pairs meet after specific time
REAL meet(mObj2D *, mObj2D *, mObj2D *, mObj2D *, REAL);
REAL meet(mObj2D *, mObj2D *, mObj2D *, mObj2D *, REAL, short, short);

// compute when two points collide after specific time
REAL bounceTime(mObj2D *, mObj2D*, REAL, REAL);

// find smallest root larger than t of equation a*x*x+b*x+c = 0
// Why I need so many solve's? - To couple with numerical degeneracy
REAL solve(REAL a, REAL b, REAL c, REAL t);
REAL solve(REAL a, REAL b, REAL c, REAL t, short nr);
REAL solve(REAL a, REAL b, REAL c, REAL t, short* nnr);
REAL solve(REAL a, REAL b, REAL c, REAL t, short nr, short* nnr);


/* moving objects */

class mObjs {
  
public:
  
  // objects
  mObj2D *mos;

  // boundary
  Boundary b;

  // queue to process event that two points cross in some direction
  eventQueue1 *queue1[3];

  // queue to process event that two pairs of points have same distance
  eventQueue2 *queue2;

  // queue to process event that one ball is bounced by wall
  eventQueue3 *queue3;

  // time to record the collision time of current closest pair
  REAL bT;

  // number of objects
  int nofObjs;

  // the index of cloest pair
  int clo1, clo2;

  // radius of balls - each ball has same radius
  REAL rr;

  mObjs() { nofObjs = 0;}
  mObjs(int);
  mObjs(int,REAL);
  mObjs(FILE* );
  ~mObjs();

  void setboundary(REAL,REAL,REAL,REAL);

  void init(void);
  void init(short);

  void setlcand(short);
  void setlcand(void);

  void reset(int );

  void generate(REAL,REAL);
  int gen_nocollision(REAL, REAL, int);
  int collide(P2,int); 
  initiate(short);

  update(flightPlan2 nfp, int obj) { return(mos[obj].update(nfp)); }

  // brute forcedly compute the closest pair at specific time
  REAL bclosest(mObj2D**, mObj2D**, REAL);

  // get the closest pair based on the current queue
  REAL getcpair(mObj2D**, mObj2D**, short, REAL);

  // process critical events
  REAL processevent(mObj2D**, mObj2D**, REAL *, REAL *);

  // fetch the time when next critical event will happen
  REAL nextcritical(void);
  REAL nextcritical(mObj2D **, mObj2D **, short *, short *);

  // process the event that two points' ranks exchange
  void exchange(mObj2D*,mObj2D*,short,REAL);

  // process the event that one point come in(or out) other's cone
  void comeInU(mObj2D*,mObj2D*,short, REAL);
  void comeInB(mObj2D*,mObj2D*,short, REAL);
  void comeOutU(mObj2D*,mObj2D*,short, REAL);
  void comeOutB(mObj2D*,mObj2D*,short, REAL);  
  void comeInUp(mObj2D*,mObj2D*,short, REAL);
  void comeInBelow(mObj2D*,mObj2D*,short, REAL);
  void comeOutUp(mObj2D*,mObj2D*,short, REAL);
  void comeOutBelow(mObj2D*,mObj2D*,short, REAL);

  // process the event that one point splits the other's cone
  void shed(mObj2D*,mObj2D*,short, REAL);

  void update(mObj2D*, mObj2D*, short, REAL);
  void update(mObj2D*, short, REAL);

  void reschedule(mObj2D*, REAL);
  bounced(mObj2D*, REAL, short);
  bounced(mObj2D*, mObj2D*, REAL);


  // for drawing the picture
  void drawBoundary(void) { draw(b,rr); }
  void drawCones(REAL,short);
  void drawParent(int,REAL,short);
  void drawUpEdge(int,REAL,short);
  void drawBelowEdge(int,REAL,short);
  savedObjs* save(REAL);
  void save(FILE*);
  void drawSaved(savedObjs*, savedObj*);
  void drawSaved(savedObj*);
  void drawSaved(savedObjs*);
};

// build up the tournament tree
void buildup(innerNode *ia, int lsize);

// update tournament tree bottom up
void updateup(innerNode *ia, int ind, int lszie);


// event queue to handle event of type 1  when two points cross 
// along some direction

class eventQueue1 {

public:

  innerNode *inode;
  mObj1D **lnode;
  int qsize,lsize;
  eventQueue1(void) { lsize = 0; qsize = 0; }
  eventQueue1(int); 
  eventQueue1(mObjs*, short);
  ~eventQueue1(void);
  set(mObjs *,short);
  set(mObjs *,short,REAL);
  REAL extractMin();
  REAL extractMin(mObj2D **, mObj2D **);
  void reschedule(int,REAL);
  short process(void);
  short process(short);
  short process(short, REAL);
  void update(int);
};

// event queue to handle event of type 2 when two pairs of points
// have same distance

class eventQueue2 {

public:

  mObj2D **lnodeT;
  innerNode *inodeT;
  int sizeTi,sizeTl,sizeQ;
  innerNode *inodeQ;
  eventQueue2(void) { sizeTi=0; sizeTl=0; sizeQ = 0;}
  eventQueue2(mObjs*);
  ~eventQueue2(void);
  void init(mObjs*);
  void init(mObjs*,REAL);
  void process(mObjs*);
  void update(mObjs *, int, short, REAL);
  REAL extractMin(void);
  REAL extractMin(mObjs *, mObj2D **, mObj2D **, short *);
  REAL extractMinDist(void);
  REAL extractMinPair(mObjs *, mObj2D **, mObj2D **);
};

// event queue to handle event of type 3 when the ball collides with
// boundary

class eventQueue3 {

public:
  
  innerNode *inode;
  short *lnode;
  int sizei,sizel;
  Boundary *b;
  eventQueue3(void) {sizei=0;sizel=0;}
  eventQueue3(mObjs*);
  eventQueue3(mObjs*, REAL);
  ~eventQueue3(void);
  void init(mObjs*);
  REAL extractMin(void);
  REAL extractMin(mObjs *, mObj2D **, short *);
  void update(mObjs *, int, REAL);
};

#endif

